
import sys

from twisted.python import log
from twisted.words.protocols.jabber import component
from twisted.words.protocols.jabber.jid import JID
from twisted.words.protocols.jabber.client import IQ
from twisted.words.xish import domish

import ns
import util
from model import Activity
from util import xpath_query, make_iq_error, make_iq_result, elem, elem_iq

valid_types = ['str', 'int', 'uint', 'bool', 'bytes']

class PropertyTypeError(Exception):
    def __init__(self, type, ns):
        self.type = type
        self.ns = ns

    def __str__(self):
        return "Invalid property type '%s'" % self.type

    def iq_error(self, stream, request):
        reply = make_iq_error(stream, request)
        error = reply.firstChildElement()
        e = error.addElement((self.ns, "invalid-property"))

        return reply

def parse_properties(elems):
    properties = {}

    for node in xpath_query('/*/property', elems):
        type = node.getAttribute('type')
        name = node.getAttribute('name')
        value = None

        for child in node.children:
            if isinstance(child, unicode) or isinstance(child, str):
                value = child
                break

        if type is None or name is None or value is None:
            continue

        if type not in valid_types:
            raise PropertyTypeError(type, elems.uri)

        if type == 'bool' and value not in ['1', '0', 'true', 'false']:
            raise PropertyTypeError(type, elems.uri)

        properties[name] = (type, value)

    return properties

def properties_to_xml(uri, properties):
    props_node = domish.Element((uri, 'properties'))

    for key, (type, value) in properties.iteritems():
        property = props_node.addElement((None, 'property'))
        property['type'] = type
        property['name'] = key
        property.addContent(value)

    return props_node

class ViewHandler(object):
    __slots__ = ('gadget', 'contact', 'id')

    def __init__(self, gadget, contact, id):
        self.gadget = gadget
        self.contact = contact
        self.id = id

    def _make_change_message(self):
        return elem('message', to=self.contact, type='notice',
                from_=self.gadget.parent.jabberId)(
                    elem('amp', xmlns=ns.AMP)(
                        elem('rule', condition='deliver-at', value='stored',
                            action='error')))

    def buddy_changed(self, buddy):
        message = self._make_change_message()
        change = message.addElement((ns.BUDDY, 'change'))
        change['jid'] = buddy.jid
        change['id'] = self.id

        change.addChild(properties_to_xml(ns.BUDDY_PROP, buddy.properties))

        self.gadget.send(message)

    def buddy_current_activity_changed(self, buddy):
        message = self._make_change_message()
        change = message.addElement((ns.BUDDY, 'change'))
        change['jid'] = buddy.jid
        change['id'] = self.id

        activity = change.addElement((ns.CURRENT_ACTIVITY, 'activity'))
        if buddy.current_activity is not None:
           activity['room'], activity['id'] = buddy.current_activity.room, buddy.current_activity.id
        else:
           activity['room'], activity['id'] = '', ''

        self.gadget.send(message)

class ActivityViewHandler(ViewHandler):
    def activity_found(self, activity):
        message = self._make_change_message()
        added = message.addElement((ns.ACTIVITY, 'added'))
        added['id'] = self.id

        added.addChild(self.gadget.activity_to_xml(activity))

        self.gadget.send(message)

    def activity_gone(self, activity):
        message = self._make_change_message()
        added = message.addElement((ns.ACTIVITY, 'removed'))
        added['id'] = self.id
        added.addChild(self.gadget.activity_to_xml(activity, False))
        self.gadget.send(message)

    def member_added(self, activity, member):
        message = self._make_change_message()
        activity_elem = self.gadget.activity_to_xml(activity, False)
        message.addChild(activity_elem)
        activity_elem['view'] = self.id
        joined = activity_elem.addElement((None, 'joined'))
        joined['jid'] = member

        if not self.gadget.model.has_buddy(member):
            # This can happen if a contact joins the activity room but is not
            # subscribed to Gadget. Then Gadget doesn't know his properties.
            log.msg("unknown member %s; can't add their properties"
                    "to query reply" % member)
        else:
            buddy = self.gadget.model.buddy_by_id(member)
            joined.addChild(properties_to_xml(ns.BUDDY_PROP, buddy.properties))

        self.gadget.send(message)

    def member_removed(self, activity, member):
        message = self._make_change_message()
        activity_elem = self.gadget.activity_to_xml(activity, False)
        message.addChild(activity_elem)
        activity_elem['view'] = self.id
        left = activity_elem.addElement((None, 'left'))
        left['jid'] = member
        self.gadget.send(message)

class BuddyViewHandler(ViewHandler):
    def buddy_added(self, buddy):
        message = self._make_change_message()
        added = message.addElement((ns.BUDDY, 'added'))
        added['id'] = self.id
        buddy_elem = added.addElement((ns.BUDDY, 'buddy'))
        buddy_elem['jid'] = str(buddy.jid)

        buddy_elem.addChild(properties_to_xml(ns.BUDDY_PROP, buddy.properties))

        self.gadget.send(message)

    def buddy_removed(self, buddy):
        message = self._make_change_message()
        removed = message.addElement((ns.BUDDY, 'removed'))
        removed['id'] = self.id
        buddy_elem = removed.addElement((ns.BUDDY, 'buddy'))
        buddy_elem['jid'] = str(buddy.jid)

        self.gadget.send(message)

class GadgetService(component.Service):
    debug = False

    features = [ns.BUDDY, ns.ACTIVITY, "%s+notify" % ns.BUDDY_PROP,
            "%s+notify" % ns.CURRENT_ACTIVITY,
            "%s+notify" % ns.NICK]
    identities = [('collaboration', 'gadget', 'OLPC Gadget')]

    def __init__(self, model, roster):
        self.model = model
        self.nickname = 'inspector'
        self.roster = roster
        # muc jid -> Activity
        self.pending_activities = {}
        # muc jid -> Activity
        self.joining_activities = {}
        # Owner JID -> (view id -> ActivityViewHandler)
        self.activity_views = {}
        # Owner JID -> (view id -> BuddyViewHandler)
        self.buddy_views = {}
        self.iq_dispatch = {
            (ns.BUDDY, 'query'): self.iq_buddy_query,
            (ns.BUDDY, 'view'): self.iq_buddy_view,
            (ns.ACTIVITY, 'query'): self.iq_activity_query,
            (ns.ACTIVITY, 'view'): self.iq_activity_view,
            (ns.DISCO_INFO, 'query'): self.iq_disco_info,
            }
        self.message_dispatch = {
            ('normal', ns.ACTIVITY, 'close'):
                self.message_close_activity_view,
            ('normal', ns.ACTIVITY_PROP, 'properties'):
                self.message_pre_invite_activity_properties,
            ('normal', ns.BUDDY, 'close'):
                self.message_close_buddy_view,
            ('normal', ns.MUC_USER, 'x'):
                self.message_muc,
            ('normal', ns.PUBSUB_EVENT, 'event'):
                self.message_pubsub_notification,
            ('groupchat', ns.ACTIVITY_PROP, 'properties'):
                self.message_activity_properties,
            }

    def send(self, e):
        if self.debug:
            out = """[SEND]
            -----------------------------------
            %s
            -----------------------------------
            """ % e.toXml()
            log.msg(out)

        if isinstance(e, IQ):
            e.send()
        else:
            component.Service.send(self, e)

    def componentConnected(self, xmlstream):
        log.msg('connected')

        self.xmlstream = xmlstream
        xmlstream.addObserver('//iq', self.iq)

        xmlstream.addObserver('//message', self.message)
        xmlstream.addObserver('//presence', self.presence)

        if self.debug:
            self.xmlstream.setDispatchFn(self.onElement)

        for jid in self.roster:
            self.send_presence(self.parent.jabberId, jid)

    def onElement(self, e):
        out = """[RECV]
        -----------------------------------
        %s
        -----------------------------------
        """ % e.toXml()
        log.msg(out)

        self.xmlstream.onElement(e)

    def componentDisconnected(self):
        self.xmlstream = None

    def iq(self, stanza):
        type = stanza.getAttribute('type')

        if type != 'get':
            return

        if None in (stanza.getAttribute('to'),
                    stanza.getAttribute('from'),
                    stanza.getAttribute('id')):
            return

        child = stanza.firstChildElement()

        if child is not None:
            f = self.iq_dispatch.get((child.uri, child.name))

            if f is not None:
                result = f(stanza)
                assert result
                self.send(result)
                return

        result = make_iq_error(self.xmlstream, stanza)
        error = result.firstChildElement()
        br = error.addElement((ns.XMPP_STANZAS, "feature-not-implemented"))
        self.send(result)

    def parse_buddy_query(self, stanza):
        result = make_iq_result(self.xmlstream, stanza)
        _query = stanza.firstChildElement()
        criteria = {}

        buddy_nodes = xpath_query('/*/buddy', _query)
        if buddy_nodes != []:
            elem = buddy_nodes[0]
            # jid search
            jid = elem.getAttribute('jid')
            if jid is not None:
                criteria['jid'] = jid

            # properties search
            for e in xpath_query(
                    '/buddy/properties[@xmlns="%s"]' % ns.BUDDY_PROP, elem):
                try:
                    criteria['properties'] = parse_properties(e)
                except PropertyTypeError, e:
                    raise ValueError(e.iq_error(self.xmlstream, stanza))

            # alias search
            alias = elem.getAttribute('alias')
            if alias is not None:
                criteria['alias'] = alias

        return criteria

    def reply_to_buddy_query(self, stanza, results):
        result = make_iq_result(self.xmlstream, stanza)
        query = result.firstChildElement()

        for buddy in results:
            buddy_elem = query.addElement((ns.BUDDY, 'buddy'))
            buddy_elem['jid'] = buddy.jid

            buddy_elem.addChild(properties_to_xml(ns.BUDDY_PROP, buddy.properties))

        return result

    def iq_buddy_query(self, stanza):
        try:
            criteria = self.parse_buddy_query(stanza)
        except ValueError, e:
            return e.args[0]

        results = self.model.buddy_search(criteria)
        return self.reply_to_buddy_query(stanza, results)

    def iq_buddy_view(self, stanza):
        from_ = stanza['from']
        # remove the resource
        jid = JID(from_).userhost()
        view_node = stanza.firstChildElement()
        id = view_node.getAttribute('id')

        if id is None:
            # missing id attribute
            result = make_iq_error(self.xmlstream, stanza)
            error = result.firstChildElement()
            error.addElement((ns.BUDDY, "invalid-property"))
            return result

        size = view_node.getAttribute('size')
        if size is None:
            # missing size attribute
            result = make_iq_error(self.xmlstream, stanza)
            error = result.firstChildElement()
            error.addElement((ns.BUDDY, "invalid-property"))
            return result

        try:
            view_size = int(size)
        except ValueError:
            # invalid size attribute
            result = make_iq_error(self.xmlstream, stanza)
            error = result.firstChildElement()
            error.addElement((ns.BUDDY, "invalid-property"))
            return result

        views = self.buddy_views.get(jid)
        if views is None:
            self.buddy_views[jid] = {}

        elif views.get(id) is not None:
            result = make_iq_error(self.xmlstream, stanza)
            error = result.firstChildElement()
            error.addElement((ns.BUDDY, "already-exists"))
            return result

        try:
            criteria = self.parse_buddy_query(stanza)
        except ValueError, e:
            return e.args[0]

        handler = BuddyViewHandler(self, jid, id)

        view = self.model.buddy_view(handler, view_size, criteria)

        self.buddy_views[jid][id] = view
        return self.reply_to_buddy_query(stanza, view.results)

    def parse_activity_query(self, stanza):
        _query = stanza.firstChildElement()
        criteria = {}

        activity_nodes = xpath_query('/*/activity', _query)
        if activity_nodes != []:
            elem = activity_nodes[0]
            buddies = None
            props = None

            # room search
            if elem.getAttribute('room') is not None:
                criteria['room'] = elem.getAttribute('room')

            for e in xpath_query(
                    '/activity/properties[@xmlns="%s"]' % ns.ACTIVITY_PROP,
                    elem):
                # properties search
                try:
                    criteria['properties'] = parse_properties(e)
                except PropertyTypeError, e:
                    raise ValueError(e.iq_error(self.xmlstream, stanza))

            for e in xpath_query('/activity/buddy', elem):
                if criteria.get('participants') is None:
                    criteria['participants'] = [e['jid']]
                else:
                    criteria['participants'].append(e['jid'])

        return criteria

    def activity_to_xml(self, activity, write_details=True):
        activity_elem = domish.Element((ns.ACTIVITY, 'activity'))
        activity_elem['id'] = str(activity.id)
        activity_elem['room'] = activity.room

        if not write_details:
            return activity_elem

        activity_elem.addChild(properties_to_xml(ns.ACTIVITY_PROP, activity.properties))

        for member in activity.members:
            member_elem = activity_elem.addElement('buddy')
            member_elem['jid'] = member

            if not self.model.has_buddy(member):
                log.msg("unknown member %s; can't add their properties"
                        "to query reply" % member)
                continue

            buddy = self.model.buddy_by_id(member)

            # FIXME: We should send buddy properties only if the user
            # doesn't know this JID yet (he doesn't have any open view
            # containing it).
            member_elem.addChild(properties_to_xml(ns.BUDDY_PROP, buddy.properties))

        return activity_elem

    def reply_to_activity_query(self, stanza, results):
        result = make_iq_result(self.xmlstream, stanza)
        query = result.firstChildElement()

        for activity in results:
            query.addChild(self.activity_to_xml(activity))
 
        return result

    def iq_activity_query(self, stanza):
        try:
            criteria = self.parse_activity_query(stanza)
        except ValueError, e:
            return e.args[0]

        results = self.model.activity_search(criteria)
        return self.reply_to_activity_query(stanza, results)

    def iq_activity_view(self, stanza):
        from_ = stanza['from']
        # remove the resource
        jid = JID(from_).userhost()
        view_node = stanza.firstChildElement()
        id = view_node.getAttribute('id')

        if id is None:
            # missing id attribute
            result = make_iq_error(self.xmlstream, stanza)
            error = result.firstChildElement()
            error.addElement((ns.ACTIVITY, "invalid-property"))
            return result

        size = view_node.getAttribute('size')
        if size is None:
            # missing size attribute
            result = make_iq_error(self.xmlstream, stanza)
            error = result.firstChildElement()
            error.addElement((ns.ACTIVITY, "invalid-property"))
            return result

        try:
            view_size = int(size)
        except ValueError:
            # invalid size attribute
            result = make_iq_error(self.xmlstream, stanza)
            error = result.firstChildElement()
            error.addElement((ns.ACTIVITY, "invalid-property"))
            return result

        views = self.activity_views.get(jid)
        if views is None:
            self.activity_views[jid] = {}

        elif views.get(id) is not None:
            result = make_iq_error(self.xmlstream, stanza)
            error = result.firstChildElement()
            error.addElement((ns.ACTIVITY, "already-exists"))
            return result

        try:
            criteria = self.parse_activity_query(stanza)
        except ValueError, e:
            return e.args[0]

        handler = ActivityViewHandler(self, jid, id)

        view = self.model.activity_view(handler, view_size, criteria)

        self.activity_views[jid][id] = view
        return self.reply_to_activity_query(stanza, view.results)

    def iq_disco_info(self, stanza):
        result = make_iq_result(self.xmlstream, stanza)
        query = result.firstChildElement()
        for category, type, name in self.identities:
            id = query.addElement('identity')
            id['category'] = category
            id['type'] = type
            id['name'] = name

        for feat in self.features:
            feature = query.addElement('feature')
            feature['var'] = feat

        return result

    def message(self, stanza):
        type = stanza.getAttribute('type', 'normal')

        if type == 'chat':
            return

        for elem in stanza.elements():
            f = self.message_dispatch.get((type, elem.uri, elem.name))

            if f is not None:
                f(stanza, elem)
                return

        log.msg('got unhandled <message>')

    def message_muc(self, stanza, x):
        for elem in x.elements():
            if elem.uri == ns.MUC_USER and elem.name == 'invite':
                self.message_muc_invite(stanza, elem)

    def create_room(self, jid, id, properties):
        # FIXME: rename
        if jid in self.pending_activities or \
            self.model.activity_by_room(jid) is not None:
            # We already have an activity by this name.
            return

        # Create a record for the activity. If we get an invitation, we'll try
        # to join it.
        activity = Activity(id, jid, properties)
        self.pending_activities[jid] = activity

    def message_pre_invite_activity_properties(self, stanza, properties_elem):
        # XXX Restrictions on from address?

        try:
            properties = parse_properties(properties_elem)
        except PropertyTypeError:
            log.msg(
                "ignored activity properties message from %s containing an "
                "invalid property" % stanza['from'])
            return

        activity_id = properties_elem.getAttribute('activity')
        room_jid = properties_elem.getAttribute('room')

        if not (properties and activity_id and room_jid):
            return

        self.create_room(room_jid, activity_id, properties)

    def message_muc_invite(self, stanza, invite):
        to = stanza.getAttribute('to')
        from_ = stanza.getAttribute('from')
        invite_from = invite.getAttribute('from')

        if None in (to, from_, invite_from):
            return

        self.invited(to, from_)

    def invited(self, to, room_jid):
        log.msg('invited to %s' % room_jid)
        activity = self.joining_activities.get(room_jid)

        if activity is not None:
            log.msg(
                'ignoring an invite to room we\'re already joining')
            return

        activity = self.pending_activities.get(room_jid)

        if not activity:
            log.msg(
                'ignoring an invite to room we didn\'t receive a property ' +
                'notification for or that we\'ve already joined')
            return

        del self.pending_activities[room_jid]
        self.joining_activities[room_jid] = activity

        presence = domish.Element((ns.CLIENT, 'presence'))
        presence['from'] = to
        j = JID(room_jid)
        j.resource = self.nickname
        presence['to'] = j.full()
        x = presence.addElement((ns.MUC_USER, 'x'))
        self.send(presence)

    def send_presence(self, _from, to):
        presence = domish.Element((None, 'presence'))
        presence['to'] = to
        presence['from'] = _from
        c = presence.addElement((ns.CAPS, 'c'))
        c['hash'] = 'sha-1'
        # FIXME: is that a right node?
        c['node'] = 'http://gadget.laptop.org'
        ver = util.hash_capabilities(self.identities, self.features)
        c['ver'] = ver
        self.send(presence)

    def presence(self, stanza):
        type = stanza.getAttribute('type')
        from_ = stanza.getAttribute('from')
        # remove the resource
        jid = JID(from_).userhost()
        to = stanza.getAttribute('to')

        if from_ is None:
            return

        for elem in stanza.elements():
            if elem.uri == ns.MUC_USER and elem.name == 'x':
                self.muc_presence(stanza)
                return

        if type == 'subscribe':
            # accept the subscription request
            result = domish.Element((None, 'presence'))
            result['type'] = 'subscribed'
            result['to'] = from_
            result['from'] = to
            self.send(result)

            log.msg('%s subscribed to gadget' % jid)

            if jid not in self.roster:
                # request subscription to the buddy
                result = domish.Element((None, 'presence'))
                result['type'] = 'subscribe'
                result['to'] = from_
                result['from'] = to
                self.send(result)

        elif type == 'subscribed':
            log.msg('subscribed to %s' % jid)

            self.roster.add(jid)
            self.model.create_buddy(jid, {})
            self.send_presence(to, from_)
            # We didn't receive PEP change notifications messages from this
            # contact so we have to request his info.
            self.request_info(from_)

        elif type == 'unsubscribed':
            log.msg('unsubscribed from %s' % jid)
            # the buddy will be removed from the model when we'll receive
            # the 'unavailable' presence
            self.roster.remove(jid)

        elif type == 'probe':
            log.msg('%s is online' % jid)

            if self.model.has_buddy(jid):
                log.msg('%s still in the model, removing them' % jid)
                self.remove_buddy(jid)

            self.model.create_buddy(jid, {})
            self.send_presence(to, from_)

        elif type == 'unavailable':
            log.msg('%s is offline' % jid)
            self.remove_buddy(jid)

    def remove_buddy(self, jid):
        try:
            self.model.buddy_remove(jid)
        except KeyError:
            pass

        # remove open views if any
        if jid in self.buddy_views:
            del self.buddy_views[jid]

        if jid in self.activity_views:
            del self.activity_views[jid]

    def muc_presence_joining(self, stanza, type, room_jid, nick):
        # Gadget is joining the room
        activity = self.joining_activities[room_jid]

        if nick == self.nickname:
            if type == 'error':
                log.msg('Error when joining room %s. Remove it.' % room_jid)
                del self.joining_activities[room_jid]
                return

            # Presence for our in-room self; we've finished joining the
            # room. Move the activity to the model.
            self.model.add_activity(activity)
            del self.joining_activities[room_jid]

            if len(activity.members) == 0:
                log.msg("room is empty, removing the activity")
                self.leave_activity(room_jid, activity.id)

            return

        if type is not None:
            log.msg("got a presence with type %s when joining room %s. Ignoring" %
                (type, room_jid))
            return

        item = xpath_query('/presence/x[@xmlns="%s"]/item' % ns.MUC_USER,
                stanza)
        if not item or item[0].getAttribute('jid') is None:
            log.msg("can't update activity members: real JID of %s is missing"
                    % nick)
            return

        real_jid = JID(item[0]['jid']).userhost()

        log.msg('room %s has member %s' % (room_jid, nick))
        activity.members.add(real_jid)

    def muc_presence(self, stanza):
        type = stanza.getAttribute('type')

        jid = JID(stanza['from'])
        room_jid = jid.userhost()
        nick = jid.resource

        log.msg('got presence for room %s from %s' % (room_jid, nick))

        # First check if we are joining this room.
        if room_jid in self.joining_activities:
            self.muc_presence_joining(stanza, type, room_jid, nick)
            return

        item = xpath_query('/presence/x[@xmlns="%s"]/item' % ns.MUC_USER,
                stanza)
        if not item or item[0].getAttribute('jid') is None:
            log.msg("can't update activity members: full jid of %s is missing"
                    % nick)
            return

        real_jid = JID(item[0]['jid']).userhost()

        activity = self.model.activity_by_room(room_jid)
        if activity is None:
            log.msg("unexpected presence message for room %s" % room_jid)
            return

        if type is None:
            log.msg('room %s has member %s' % (room_jid, nick))
            self.model.activity_add_member(activity.id, real_jid)

        elif type == 'unavailable':
            if real_jid in activity.members:
                self.model.activity_remove_member(activity.id, real_jid)

                if len(activity.members) == 0:
                    self.leave_activity(room_jid, activity.id)

    def leave_activity(self, room_jid, activity_id):
        log.msg('leaving room %s' % room_jid)

        presence = elem('presence', to='%s/%s' % (room_jid, self.nickname),
                from_=self.parent.jabberId, type='unavailable')
        self.send(presence)

        if self.model.has_activity(activity_id):
            self.model.activity_remove(activity_id)

        if room_jid in self.pending_activities:
            del self.pending_activities[room_jid]

        if room_jid in self.joining_activities:
            del self.joining_activities[room_jid]

    def message_pubsub_notification(self, stanza, event):
        items = event.firstChildElement()

        if items['node'] == ns.BUDDY_PROP:
            self.message_buddy_properties_notification(stanza, items)
        elif items['node'] == ns.CURRENT_ACTIVITY:
            self.message_buddy_current_activity_notification(stanza, items)
        elif items['node'] == ns.NICK:
            self.message_buddy_nick_notification(stanza, items)

    def message_buddy_properties_notification(self, stanza, items):
        from_ = stanza.getAttribute('from')
        # remove the resource
        jid = JID(from_).userhost()

        for properties in xpath_query(
                '/items/item/properties[@xmlns="%s"]' % ns.BUDDY_PROP, items):
            props = parse_properties(properties)

            if self.model.has_buddy(jid):
                self.model.buddy_update(jid, props)
            else:
                self.model.create_buddy(jid, props)

    def message_buddy_current_activity_notification(self, stanza, items):
        from_ = stanza.getAttribute('from')
        # remove the resource
        jid = JID(from_).userhost()

        for activity in xpath_query(
                '/items/item/activity[@xmlns="%s"]' % ns.CURRENT_ACTIVITY, items):
            room, id = activity.getAttribute('room'), activity.getAttribute('type')

            if room is None or id is None or room == '' or id == '':
                activity = None
            else:
                # FIXME: Should we look in the model if there is an Activity
                # object with the id and room?
                # What should we do if there is one Activity with the same id
                # but with a different room (or the opposite). That shouldn't
                # happen but, well...
                activity = Activity(id, room, {})

            if self.model.has_buddy(jid):
                self.model.buddy_change_current_activity(jid, activity)

    def message_activity_properties(self, stanza, properties_elem):
        id = properties_elem.getAttribute('activity')
        room_jid = properties_elem.getAttribute('room')
        jid = JID(stanza['from'])

        if room_jid != jid.userhost():
            log.msg(
                "ignored activity properties message from %s trying to "
                "change properties for another room" % stanza['from'])
            return

        try:
            properties = parse_properties(properties_elem)
        except PropertyTypeError:
            log.msg(
              "ignored activity properties message from %s containing "
              "an invalid property" % stanza['from'])
            return

        if not self.model.has_activity(id):
            log.msg(
                "ignored activity properties message from %s about an "
                "invalid activity: %s " % (stanza['from'], id))
            return

        # Is the activity private?
        if properties.get('private') in (('bool', '1'), ('bool', 'true')):
            log.msg('activity %s is not private; leaving it' % id)
            self.leave_activity(room_jid, id)
            return

        self.model.activity_update(id, properties)

    def message_close_activity_view(self, stanza, close):
        from_ = stanza['from']
        # remove the resource
        jid = JID(from_).userhost()
        id = close['id']
        if not id:
            log.msg('Ignoring activity view close stanza from %s without ID'
                    % jid)
            return

        views = self.activity_views.get(jid, {})
        if not views.has_key(id):
            log.msg('Ignoring activity view close stanza from %s, unknown ID %s'
                    % (jid, id))
            return

        del views[id]

        if len(views) == 0:
            del self.activity_views[jid]

    def message_close_buddy_view(self, stanza, close):
        from_ = stanza['from']
        # remove the resource
        jid = JID(from_).userhost()
        id = close['id']
        if not id:
            log.msg('Ignoring buddy view close stanza from %s without ID'
                    % jid)
            return

        views = self.buddy_views.get(jid, {})
        if not views.has_key(id):
            log.msg('Ignoring buddy view close stanza from %s, unknown ID %s'
                    % (jid, id))
            return

        del views[id]

        if len(views) == 0:
            del self.buddy_views[jid]

    def message_buddy_nick_notification(self, stanza, items):
        from_ = stanza.getAttribute('from')
        # remove the resource
        jid = JID(from_).userhost()

        for nick_node in xpath_query(
                '/items/item/nick[@xmlns="%s"]' % ns.NICK, items):
            nick = nick_node.children[0]
            if nick == '':
                return

            if self.model.has_buddy(jid):
                self.model.buddy_change_alias(jid, nick)

    def pubsub_query(self, jid, node):
        return elem_iq(self.xmlstream, 'get', to=jid, from_=self.parent.jabberId)(
            elem(ns.PUBSUB, 'pubsub')(
                elem('items', node=node)))

    def got_buddy_prop_reply(self, reply, jid):
        if jid != reply.getAttribute('from'):
            return False

        items = xpath_query('/iq/pubsub/items[@node="%s"]' % ns.BUDDY_PROP, reply)
        if len(items) == 0:
            return True

        self.message_buddy_properties_notification(reply, items[0])
        return True

    def got_current_activity_reply(self, reply, jid):
        if jid != reply.getAttribute('from'):
            return False

        items = xpath_query('/iq/pubsub/items[@node="%s"]' % ns.CURRENT_ACTIVITY, reply)
        if len(items) == 0:
            return True

        self.message_buddy_current_activity_notification(reply, items[0])
        return True

    def got_nick_reply(self, reply, jid):
        if jid != reply.getAttribute('from'):
            return False

        items = xpath_query('/iq/pubsub/items[@node="%s"]' % ns.NICK, reply)
        if len(items) == 0:
            return True

        self.message_buddy_nick_notification(reply, items[0])
        return True

    def request_info(self, jid):
        # request buddy properties
        query = self.pubsub_query(jid, ns.BUDDY_PROP)
        query.addCallback(lambda e: self.got_buddy_prop_reply(e, jid))
        self.send(query)

        # request current activity
        query = self.pubsub_query(jid, ns.CURRENT_ACTIVITY)
        query.addCallback(lambda e: self.got_current_activity_reply(e, jid))
        self.send(query)

        # request nick
        query = self.pubsub_query(jid, ns.NICK)
        query.addCallback(lambda e: self.got_nick_reply(e, jid))
        self.send(query)
