
from twisted.internet import defer, reactor
from twisted.words.protocols.jabber.client import IQ

from model import Buddy, Activity
from test_component import TestCase
from util import xpath_query, elem, elem_find_children

import ns

def parse_properties(node):
    properties = {}

    for child in elem_find_children(node, ns.BUDDY_PROP, 'properties'):
        for property in elem_find_children(child, ns.BUDDY_PROP, 'property'):
            name = property['name']
            type = property['type']
            value = property.children[0]
            properties[name] = (type, value)

    return properties

def parse_buddies_result(stanza):
    assert stanza.getAttribute('type') == 'result'
    query = stanza.firstChildElement()
    assert query.name in ['query', 'view']
    buddies = []

    for child in query.elements():
        if child.name == 'buddy':
            jid = child.getAttribute('jid')
            assert jid

            properties = parse_properties(child)
            buddies.append(Buddy(jid, properties))

    return sorted(buddies)

class BuddyIqTest(TestCase):
    """Empty buddy query with no results."""

    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        iq = IQ(self.server, "get")
        iq['from'] = 'bob@foo.org'
        iq['to'] = 'gadget.bar.org'
        query = iq.addElement((ns.BUDDY, 'query'))
        iq.send()

        return self.done

    def iq(self, stanza):
        self.assertEquals([], parse_buddies_result(stanza))
        self.done.callback(None)

class BuddyIqTest2(TestCase):
    """Empty buddy query with two results."""

    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        model = self.service.model
        model.create_buddy('foo', {})
        model.create_buddy('bar', {})

        iq = IQ(self.server, "get")
        iq['from'] = 'bob@foo.org'
        iq['to'] = 'gadget.bar.org'
        query = iq.addElement((ns.BUDDY, 'query'))
        iq.send()

        return self.done

    def iq(self, stanza):
        self.assertEquals([Buddy('bar'), Buddy('foo')],
            parse_buddies_result(stanza))
        self.done.callback(None)

class BuddyIqTest3(TestCase):
    """JID based buddy query with one result"""

    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        model = self.service.model
        model.create_buddy('alice@foo.org',
            {'color': ('str', '#005FE4,#00A0FF')})
        model.create_buddy('marc@foo.org', {})

        iq = IQ(self.server, "get")
        iq['from'] = 'bob@foo.org'
        iq['to'] = 'gadget.bar.org'
        query = iq.addElement((ns.BUDDY, 'query'))
        buddy = query.addElement((None, 'buddy'))
        buddy['jid'] = 'alice@foo.org'
        iq.send()

        return self.done

    def iq(self, stanza):
        self.assertEquals(
            [Buddy('alice@foo.org', {'color': ('str', '#005FE4,#00A0FF')})],
            parse_buddies_result(stanza))
        self.done.callback(None)

class BuddyIqTest4(TestCase):
    """properties based buddy query with one result"""

    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        model = self.service.model
        model.create_buddy('alice@foo.org',
            {'color': ('str', '#005FE4,#00A0FF')})
        model.create_buddy('marc@foo.org', {})

        iq = IQ(self.server, "get")
        iq['from'] = 'bob@foo.org'
        iq['to'] = 'gadget.bar.org'
        query = iq.addElement((ns.BUDDY, 'query'))
        buddy = query.addElement((None, 'buddy'))
        properties = buddy.addElement((ns.BUDDY_PROP, 'properties'))
        property = properties.addElement((None, 'property'))
        property['type'] = 'str'
        property['name'] = 'color'
        property.addContent('#005FE4,#00A0FF')
        iq.send()

        return self.done

    def iq(self, stanza):
        self.assertEquals(
            [Buddy('alice@foo.org', {'color': ('str', '#005FE4,#00A0FF')})],
            parse_buddies_result(stanza))
        self.done.callback(None)

class BuddyIqInvalidPropertyTest(TestCase):
    """buddy query with an invalid property"""

    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        iq = IQ(self.server, "get")
        iq['from'] = 'bob@foo.org'
        iq['to'] = 'gadget.bar.org'
        query = iq.addElement((ns.BUDDY, 'query'))
        buddy = query.addElement((None, 'buddy'))
        properties = buddy.addElement((ns.BUDDY_PROP, 'properties'))
        property = properties.addElement((None, 'property'))
        property['type'] = 'wtf'
        property['name'] = 'nothing'
        property.addContent('hey')
        iq.send()

        return self.done

    def iq(self, stanza):
        assert stanza.getAttribute('type') == 'error'
        error = stanza.firstChildElement()
        assert error.name == 'error'
        e = error.firstChildElement()
        assert e.name == 'invalid-property'
        assert e.uri == ns.BUDDY_PROP

        self.done.callback(None)

class BuddyIqRandomQuery(TestCase):
    """buddy random view query"""

    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        model = self.service.model
        model.create_buddy('alice@foo.org', {})
        model.create_buddy('bob@foo.org', {})
        model.create_buddy('charles@foo.org', {})
        model.create_buddy('max@foo.org', {})

        iq = IQ(self.server, "get")
        iq['from'] = 'bob@foo.org/Resource'
        iq['to'] = 'gadget.bar.org'
        view = iq.addElement((ns.BUDDY, 'view'))
        view['id'] = 'query1'
        view['size'] = '3'
        iq.send()

        return self.done

    def iq(self, stanza):
        views = self.service.buddy_views['bob@foo.org']
        view = views['query1']
        assert view.size == 3

        buddies = parse_buddies_result(stanza)
        assert len(buddies) == 3

        self.done.callback(None)

class BuddyViewTest(TestCase):
    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        model = self.service.model
        model.create_buddy('alice@foo.org', {'color': ('str', 'red')})
        model.create_buddy('marc@foo.org', {})

        iq = IQ(self.server, "get")
        iq['from'] = 'bob@foo.org'
        iq['to'] = 'gadget.bar.org'
        view = iq.addElement((ns.BUDDY, 'view'))
        view['id'] = 'view1'
        view['size'] = '20'
        buddy = view.addElement((None, 'buddy'))
        properties = buddy.addElement((ns.BUDDY_PROP, 'properties'))
        property = properties.addElement((None, 'property'))
        property['type'] = 'str'
        property['name'] = 'color'
        property.addContent('red')
        iq.send()

        return self.done

    def iq(self, stanza):
        views = self.service.buddy_views['bob@foo.org']
        view = views['view1']
        assert view.size == 20

        self.assertEquals(
            [Buddy('alice@foo.org', {'color': ('str', 'red')})],
            parse_buddies_result(stanza))

        self.server.addOnetimeObserver('//message', self.message)

        # add a matching buddy
        self.service.model.create_buddy('lucien@foo.org',
                {'color': ('str', 'red')})

    def message(self, stanza):
        assert stanza['to'] == 'bob@foo.org'
        assert stanza['from'] == 'gadget.localhost'
        added = xpath_query('/message/added[@xmlns="%s"]'
                % ns.BUDDY, stanza)[0]
        assert added['id'] == 'view1'
        buddy = added.firstChildElement()
        assert buddy.name == 'buddy'
        assert buddy['jid'] == 'lucien@foo.org'
        self.assertEquals({'color': ('str', 'red')}, parse_properties(buddy))

        self.server.addOnetimeObserver('//message', self.message2)

        # alice changed her properties and doesn't match anymore
        self.service.model.buddy_update('alice@foo.org',
                {'color': ('str', 'blue')})

    def message2(self, stanza):
        assert stanza['to'] == 'bob@foo.org'
        assert stanza['from'] == 'gadget.localhost'
        removed = xpath_query('/message/removed[@xmlns="%s"]'
                % ns.BUDDY, stanza)[0]
        assert removed['id'] == 'view1'
        buddy = removed.firstChildElement()
        assert buddy.name == 'buddy'
        assert buddy['jid'] == 'alice@foo.org'

        self.server.addOnetimeObserver('//message', self.message3)
        # Lucien changed one of his property but still match
        self.service.model.buddy_update('lucien@foo.org',
                {'color': ('str', 'red'), 'name': ('str', 'Lucien')})

    def message3(self, stanza):
        assert stanza['to'] == 'bob@foo.org'
        assert stanza['from'] == 'gadget.localhost'
        change = xpath_query('/message/change[@xmlns="%s"]'
                % ns.BUDDY, stanza)[0]
        assert change['jid'] == 'lucien@foo.org'
        assert change['id'] == 'view1'

        props = parse_properties(change)
        self.assertEquals({'color': ('str', 'red'), 'name': ('str', 'Lucien')},
                props)

        self.server.addOnetimeObserver('//message', self.message4)
        # Lucien changed his current activity
        self.service.model.buddy_change_current_activity('lucien@foo.org',
            Activity('room1', 'id1', {}))

    def message4(self, stanza):
        assert stanza['to'] == 'bob@foo.org'
        assert stanza['from'] == 'gadget.localhost'
        change = xpath_query('/message/change[@xmlns="%s"]'
                % ns.BUDDY, stanza)[0]
        assert change['jid'] == 'lucien@foo.org'
        assert change['id'] == 'view1'
        activity = xpath_query('/change/activity[@xmlns="%s"]'
                % ns.CURRENT_ACTIVITY, change)[0]
        activity['room'] == 'room1'
        activity['id'] == 'id1'

        self.server.addOnetimeObserver('//message', self.message5)
        # Lucien doesn't have current activity anymore
        self.service.model.buddy_change_current_activity('lucien@foo.org',
            None)

    def message5(self, stanza):
        assert stanza['to'] == 'bob@foo.org'
        assert stanza['from'] == 'gadget.localhost'
        change = xpath_query('/message/change[@xmlns="%s"]'
                % ns.BUDDY, stanza)[0]
        assert change['jid'] == 'lucien@foo.org'
        assert change['id'] == 'view1'
        activity = xpath_query('/change/activity[@xmlns="%s"]'
                % ns.CURRENT_ACTIVITY, change)[0]
        activity['room'] == ''
        activity['id'] == ''

        # close the view
        message = elem('message', from_='bob@foo.org/Telepathy', to='gadget.localhost')(
            elem('close', xmlns=ns.BUDDY, id='view1'))

        self.server.send(message)
        reactor.callLater(0, self.later)

    def later(self):
        assert self.service.buddy_views == {}
        # view was removed from the model too
        assert len(self.model.buddy_views) == 0
        self.done.callback(None)

class BuddyIqViewNoId(TestCase):
    """buddy random view query without ID"""
    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        iq = IQ(self.server, "get")
        iq['from'] = 'bob@foo.org'
        iq['to'] = 'gadget.bar.org'
        view = iq.addElement((ns.BUDDY, 'view'))
        view['size'] = '3'
        iq.send()

    def iq(self, stanza):
        assert stanza.getAttribute('type') == 'error'
        assert len(xpath_query('/iq/error/invalid-property[@xmlns="%s"]' %
            ns.BUDDY, stanza)) == 1

        self.done.callback(None)

class BuddyIqViewExist(TestCase):
    """buddy random view query using an existing ID"""

    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        # create a fake view with 'query1' as ID
        self.service.buddy_views['bob@foo.org'] = {}
        self.service.buddy_views['bob@foo.org']['query1'] = True
        iq = IQ(self.server, "get")
        iq['from'] = 'bob@foo.org'
        iq['to'] = 'gadget.bar.org'
        view = iq.addElement((ns.BUDDY, 'view'))
        view['id'] = 'query1'
        view['size'] = '20'
        iq.send()

        return self.done

    def iq(self, stanza):
        assert stanza['from'] == 'gadget.bar.org'
        assert stanza.getAttribute('type') == 'error'
        assert len(xpath_query('/iq/error/already-exists[@xmlns="%s"]' %
            ns.BUDDY, stanza)) == 1

        self.done.callback(None)

class BuddyIqAliasQuery(TestCase):
    """buddy alias view query"""
    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        model = self.service.model
        buddy = model.create_buddy('tom@foo.org', {})
        buddy.alias = 'tom'
        buddy = model.create_buddy('thomas@foo.org', {})
        buddy.alias = 'tom'
        model.create_buddy('charles@foo.org', {})

        iq = IQ(self.server, "get")
        iq['from'] = 'bob@foo.org'
        iq['to'] = 'gadget.bar.org'
        view = iq.addElement((ns.BUDDY, 'view'))
        view['id'] = 'query1'
        view['size'] = '20'
        buddy = view.addElement((None, 'buddy'))
        buddy['alias'] = 'tom'
        iq.send()

        return self.done

    def iq(self, stanza):
        self.assertEquals(
            [Buddy('thomas@foo.org', {}), Buddy('tom@foo.org')],
            parse_buddies_result(stanza))

        self.done.callback(None)

class BuddyNotInViewRemoved(TestCase):
    """A buddy which is not in a buddy view is removed from the model"""

    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        model = self.service.model
        model.create_buddy('alice@foo.org', {})
        model.create_buddy('charles@foo.org', {})

        iq = IQ(self.server, "get")
        iq['from'] = 'bob@foo.org/Resource'
        iq['to'] = 'gadget.bar.org'
        view = iq.addElement((ns.BUDDY, 'view'))
        view['id'] = 'query1'
        view['size'] = '1'
        iq.send()

        return self.done

    def iq(self, stanza):
        buddies = parse_buddies_result(stanza)
        buddy = buddies[0]

        # remove the buddy which is not in the view
        if buddy.jid == 'alice@foo.org':
            self.service.remove_buddy('charles@foo.org')
        else:
            self.service.remove_buddy('alice@foo.org')

        self.done.callback(None)

class BuddyIqViewInvalidSize(TestCase):
    """buddy view query with an invalid size attribute"""

    def runTest(self):
        self.done = defer.Deferred()
        self.server.addOnetimeObserver('//iq', self.iq)

        iq = IQ(self.server, "get")
        iq['from'] = 'bob@foo.org'
        iq['to'] = 'gadget.bar.org'
        view = iq.addElement((ns.BUDDY, 'view'))
        view['id'] = 'query1'
        view['id'] = 'foo'
        iq.send()

        return self.done

    def iq(self, stanza):
        assert stanza.getAttribute('type') == 'error'
        error = stanza.firstChildElement()
        assert error.name == 'error'
        e = error.firstChildElement()
        assert e.name == 'invalid-property'

        self.done.callback(None)
